/*
Copyright (C) 1999/2000 Hendrik Lipka.
All rights reserved.

This program is free software; you can redistribute it and/or modify it under the terms 
of the GNU General Public License as published by the Free Software Foundation; 
either version 2 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY 
WARRANTY; without even the implied warranty of MERCHANTABILITY or 
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License 
for more details. 

You should have received a copy of the GNU General Public License along with this 
program; if not, write to the Free Software Foundation, Inc., 675 Mass Ave, 
Cambridge, MA 02139, USA or see http://www.gnu.org

If you encounter problems or have a suggestion, please mail to:
ldse@hendriklipka.de
*/

package de.hendriklipka.ldse.common;

import java.io.FilterReader;
import java.io.Reader;
import java.io.IOException;

import de.hendriklipka.ldse.common.Log;

/**
  Entfernt Kommentare aus einem HTML-Stream
  State-Machine...
*/
public class HTMLCommentFilterReader extends FilterReader
{
  protected int _state=NORMAL;
  protected final static int NORMAL=0;
  protected final static int COMMENT_IN1=1;   // <
  protected final static int COMMENT_IN2=2;   // !
  protected final static int COMMENT_IN3=3;   // -
  protected final static int IN_STRING=5;
  
  protected final static int _maxPreChars=4;
  protected char[] _preBuffer=new char[_maxPreChars];
  protected int _preChars=0;
  
  protected boolean _finished=false;
  
	public HTMLCommentFilterReader(Reader in)
	{
		super(in);
	}
	
  /**
    * Read characters into a portion of an array.
    * does the filtering
    * @exception  IOException  If an I/O error occurs
  */
  public int read(char cbuf[], int off, int len) throws IOException 
  {
    int max=off+len;
    for (int i=off;i<max;i++)
    {
      int c=read();
      if (-1!=c)
        cbuf[i]=(char)c;
      else
        return -1;
    }
    return len;
  }
  
  /**
    * Read a single character.  This method will block until a character is
    * available, an I/O error occurs, or the end of the stream is reached.
    *
    * <p> Subclasses that intend to support efficient single-character input
    * should override this method.
    *
    * @return     The character read, as an integer in the range 0 to 16383
    *             (<tt>0x00-0xffff</tt>), or -1 if the end of the stream has
    *             been reached
    *
    * @exception  IOException  If an I/O error occurs
  */
  // hier muss das eigentlich parsen passieren
  public int read() throws IOException 
  {
    int i=readNext();
    if (-1==i)
      return -1;
    char c=(char)i;
    if ('<'==c)
      if ('!'==(char)readNext(0))
        if ('-'==(char)readNext(1))
          if ('-'==(char)readNext(2))
          { // Kommentaranfang gefunden -> ueberlesen
            // Anfang auslesen
            readNext(); // !
            readNext(); // -
            readNext(); // -
            while (true)
            {
              i=readNext();
              if (-1==i)
                return -1;
              if ('-'==(char)i) // Anfang vom Ende des Kommentars?
                if('-'==(char)readNext(0)) // "--" darf nur am Kommentarende auftreten, also sind wir hier sicher
                  if('>'==(char)readNext(1))
                  { // Kommentarende
                    readNext(); // -
                    readNext(); // >
                    return read();
                  }
            }
          }
    return i;
  }

  /**
    * Read characters into an array.  This method will block until some input
    * is available, an I/O error occurs, or the end of the stream is reached.
    *
    * @param       cbuf  Destination buffer
    *
    * @return      The number of bytes read, or -1 if the end of the stream
    *              has been reached
    *
    * @exception   IOException  If an I/O error occurs
  */
  public int read(char cbuf[]) throws IOException 
  {
    return read(cbuf, 0, cbuf.length);
  }

  /**
    * Skip characters.  This method will block until some characters are
    * available, an I/O error occurs, or the end of the stream is reached.
    *
    * @param  n  The number of characters to skip
    *
    * @return    The number of characters actually skipped
    *
    * @exception  IOException  If an I/O error occurs
  */
  private static final int maxSkipBufferSize = 8192;
  private char skipBuffer[] = null;
  public long skip(long n) throws IOException 
  {
    int nn = (int) Math.min(n, maxSkipBufferSize);
    synchronized (lock) 
    {
      if ((skipBuffer == null) || (skipBuffer.length < nn))
        skipBuffer = new char[nn];
      long r = n;
      while (r > 0) 
      {
        int nc = read(skipBuffer, 0, nn);
        if (nc == -1)
          break;
        r -= nc;
      }
      return n - r;
    }
  }
    
  // liefert das naechste Zeichen des Streams, wobei stets ein LookAHead
  // ueber 4 Zeichen durchgefuehrt wird
  protected int readNext() throws IOException
  {
    if (0==_preChars)
    {
      if (_finished)
        return -1;
      int i=in.read();
      if (-1==i)
        _finished=true;
      return i;
    }
    int c=_preBuffer[0];
    _preChars--;
    for (int i=0;i<_preChars;i++)
    {
      _preBuffer[i]=_preBuffer[i+1];
    }
    return c;
  }
  
  // schaut i Zeichen voraus
  // i=0 liefert das naechste zeichen, ohne es des LookAHead-Puffer zu entnehmen
  // i<_maxPreChars (4)
  protected int readNext(int i) throws IOException
  {
    if (i<_preChars)
      return _preBuffer[i];
    while (_preChars<=i)
    {
      int c=in.read();
      if (-1==c)
      {
        if (0==_preChars)
        {
          _finished=true;
          return -1;
        }
        break;
      }
      _preBuffer[_preChars++]=(char)c;
    }
    if (i<_preChars)
      return _preBuffer[i];
    else
      return -1;
  }
 
  public void close() throws IOException
  {
    in.close();
  }
  
  public boolean ready() throws IOException 
  {
    return in.ready();
  }
}